"""Define a SimpliSafe lock."""
from enum import Enum
import logging

from simplipy.entity import EntityV3

_LOGGER: logging.Logger = logging.getLogger(__name__)


class LockStates(Enum):
    """States that a lock can be in."""

    unlocked = 0
    locked = 1
    jammed = 2
    unknown = 99


SET_STATE_MAP = {LockStates.locked: "lock", LockStates.unlocked: "unlock"}


class Lock(EntityV3):
    """A lock that works with V3 systems.

    Note that this class shouldn't be instantiated directly; it will be
    instantiated as appropriate via :meth:`simplipy.API.get_systems`.
    """

    @property
    def disabled(self) -> bool:
        """Return whether the lock is disabled.

        :rtype: ``bool``
        """
        return self.entity_data["status"]["lockDisabled"]

    @property
    def lock_low_battery(self) -> bool:
        """Return whether the lock's battery is low.

        :rtype: ``bool``
        """
        return self.entity_data["status"]["lockLowBattery"]

    @property
    def pin_pad_low_battery(self) -> bool:
        """Return whether the pin pad's battery is low.

        :rtype: ``bool``
        """
        return self.entity_data["status"]["pinPadLowBattery"]

    @property
    def pin_pad_offline(self) -> bool:
        """Return whether the pin pad is offline.

        :rtype: ``bool``
        """
        return self.entity_data["status"]["pinPadOffline"]

    @property
    def state(self) -> LockStates:
        """Return the current state of the lock.

        :rtype: :meth:`simplipy.lock.LockStates`
        """
        if bool(self.entity_data["status"]["lockJamState"]):
            return LockStates.jammed

        raw_state = self.entity_data["status"]["lockState"]

        try:
            return LockStates(raw_state)
        except ValueError:
            _LOGGER.error("Unknown raw lock state: %s", raw_state)
            return LockStates.unknown

    async def _set_lock_state(self, state: LockStates) -> None:
        """Set the lock state."""
        await self._request(
            "post",
            f"doorlock/{self._system_id}/{self.serial}/state",
            json={"state": SET_STATE_MAP[state]},
        )

        self.entity_data["status"]["lockState"] = state.value

    async def lock(self) -> None:
        """Lock the lock."""
        await self._set_lock_state(LockStates.locked)

    async def unlock(self) -> None:
        """Unlock the lock."""
        await self._set_lock_state(LockStates.unlocked)



"""Define a v3 (new) SimpliSafe sensor."""
import logging
from typing import Optional

from simplipy.entity import EntityTypes, EntityV3

_LOGGER: logging.Logger = logging.getLogger(__name__)


class SensorV3(EntityV3):
    """A V3 (new) sensor.

    Note that this class shouldn't be instantiated directly; it will be
    instantiated as appropriate via :meth:`simplipy.API.get_systems`.
    """

    @property
    def trigger_instantly(self) -> bool:
        """Return whether the sensor will trigger instantly.

        :rtype: ``bool``
        """
        return self.entity_data["setting"].get("instantTrigger", False)

    @property
    def triggered(self) -> bool:
        """Return whether the sensor has been triggered.

        :rtype: ``bool``
        """
        if self.type in (
            EntityTypes.carbon_monoxide,
            EntityTypes.entry,
            EntityTypes.glass_break,
            EntityTypes.leak,
            EntityTypes.motion,
            EntityTypes.smoke,
            EntityTypes.temperature,
        ):
            return self.entity_data["status"].get("triggered", False)

        return False

    @property
    def temperature(self) -> Optional[int]:
        """Return the temperature of the sensor (as appropriate).

        If the sensor isn't a temperature sensor, an ``AttributeError`` will be raised.

        :rtype: ``int``
        """
        if self.type != EntityTypes.temperature:
            raise AttributeError("Non-temperature sensor cannot have a temperature")

        return self.entity_data["status"]["temperature"]



"""Define datetime utilities."""
from datetime import datetime

import pytz

UTC = pytz.utc


def utc_from_timestamp(timestamp: float) -> datetime:
    """Return a UTC time from a timestamp."""
    return UTC.localize(datetime.utcfromtimestamp(timestamp))



"""Define V2 and V3 SimpliSafe systems."""
import asyncio
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
import logging
from typing import Any, Callable, Coroutine, Dict, List, Optional, Set, Type, Union

from simplipy.entity import Entity, EntityTypes
from simplipy.errors import PinError, SimplipyError
from simplipy.lock import Lock
from simplipy.sensor.v2 import SensorV2
from simplipy.sensor.v3 import SensorV3
from simplipy.util.dt import utc_from_timestamp
from simplipy.util.string import convert_to_underscore

_LOGGER: logging.Logger = logging.getLogger(__name__)

VERSION_V2 = 2
VERSION_V3 = 3

EVENT_SYSTEM_NOTIFICATION = "system_notification"

CONF_DEFAULT: str = "default"
CONF_DURESS_PIN: str = "duress"
CONF_MASTER_PIN: str = "master"

DEFAULT_MAX_USER_PINS: int = 4
MAX_PIN_LENGTH: int = 4
RESERVED_PIN_LABELS: Set[str] = {CONF_DURESS_PIN, CONF_MASTER_PIN}

ENTITY_MAP: Dict[int, dict] = {
    VERSION_V2: {CONF_DEFAULT: SensorV2},
    VERSION_V3: {CONF_DEFAULT: SensorV3, EntityTypes.lock: Lock},
}


def guard_from_missing_data(default_value: Any = None):
    """Guard a missing property by returning a set value."""

    def decorator(func):
        """Decorate."""

        def wrapper(system):
            """Call the function and handle any issue."""
            try:
                return func(system)
            except KeyError as err:
                _LOGGER.warning(
                    "SimpliSafe didn't return data for property: %s", func.__name__
                )
                _LOGGER.debug(err)
                return default_value

        return wrapper

    return decorator


def create_pin_payload(pins: dict, *, version: int = VERSION_V3) -> dict:
    """Create the appropriate PIN payload for the provided version."""
    empty_user_index: int
    idx: int
    label: str
    payload: dict
    pin: str

    if version == VERSION_V2:
        payload = {
            "pins": {
                CONF_DURESS_PIN: {"value": pins.pop(CONF_DURESS_PIN)},
                "pin1": {"value": pins.pop(CONF_MASTER_PIN)},
            }
        }

        empty_user_index = len(pins)
        for idx, (label, pin) in enumerate(pins.items()):
            payload["pins"][f"pin{idx + 2}"] = {"name": label, "value": pin}

        for idx in range(DEFAULT_MAX_USER_PINS - empty_user_index):
            payload["pins"][f"pin{str(idx + 2 + empty_user_index)}"] = {
                "name": "",
                "pin": "",
            }
    else:
        payload = {
            "pins": {
                CONF_DURESS_PIN: {"pin": pins.pop(CONF_DURESS_PIN)},
                CONF_MASTER_PIN: {"pin": pins.pop(CONF_MASTER_PIN)},
                "users": {},
            }
        }

        for idx, (label, pin) in enumerate(pins.items()):
            payload["pins"]["users"][str(idx)] = {"name": label, "pin": pin}

        empty_user_index = len(pins)
        for idx in range(DEFAULT_MAX_USER_PINS - empty_user_index):
            payload["pins"]["users"][str(idx + empty_user_index)] = {
                "name": "",
                "pin": "",
            }

    return payload


def get_entity_class(
    entity_type: EntityTypes, *, version: int = VERSION_V3
) -> Type[Entity]:
    """Return the appropriate entity class based on version and entity type."""
    return ENTITY_MAP[version].get(entity_type, ENTITY_MAP[version][CONF_DEFAULT])


@dataclass(frozen=True)  # pylint: disable=too-many-instance-attributes
class SystemNotification:
    """Define a representation of a system notification."""

    notification_id: str
    text: str
    category: str
    code: str
    timestamp: datetime

    link: Optional[str] = None
    link_label: Optional[str] = None

    def __post_init__(self):
        """Run post-init initialization."""
        object.__setattr__(self, "timestamp", utc_from_timestamp(self.timestamp))


class SystemStates(Enum):
    """States that the system can be in."""

    alarm = 1
    alarm_count = 2
    away = 3
    away_count = 4
    entry_delay = 5
    error = 6
    exit_delay = 7
    home = 8
    home_count = 9
    off = 10
    unknown = 99


class System:
    """Define a system.

    Note that this class shouldn't be instantiated directly; it will be instantiated as
    appropriate via :meth:`simplipy.API.get_systems`.

    :param request: A method to make authenticated API requests.
    :type request: ``Callable[..., Coroutine]``
    :param get_subscription_data: A method to get the latest subscription data
    :type get_subscription_data: ``Callable[..., Coroutine]``
    :param location_info: A raw data dict representing the system's state and properties.
    :type location_info: ``dict``
    """

    def __init__(
        self,
        request: Callable[..., Coroutine],
        get_subscription_data: Callable[..., Coroutine],
        location_info: dict,
    ) -> None:
        """Initialize."""
        self._get_subscription_data: Callable[..., Coroutine] = get_subscription_data
        self._location_info: dict = location_info
        self._notifications: List[
            SystemNotification
        ] = self._generate_system_notification_objects()
        self._request: Callable[..., Coroutine] = request
        self._state: SystemStates = self._coerce_state_from_string(
            location_info["system"]["alarmState"]
        )
        self.locks: Dict[str, Lock] = {}
        self.sensors: Dict[str, Union[SensorV2, SensorV3]] = {}

    @property  # type: ignore
    @guard_from_missing_data()
    def address(self) -> str:
        """Return the street address of the system.

        :rtype: ``str``
        """
        return self._location_info["street1"]

    @property  # type: ignore
    @guard_from_missing_data(False)
    def alarm_going_off(self) -> bool:
        """Return whether the alarm is going off.

        :rtype: ``bool``
        """
        return self._location_info["system"]["isAlarming"]

    @property  # type: ignore
    @guard_from_missing_data()
    def connection_type(self) -> str:
        """Return the system's connection type (cell or WiFi).

        :rtype: ``str``
        """
        return self._location_info["system"]["connType"]

    @property
    def notifications(self) -> List[SystemNotification]:
        """Return the system's current messages/notifications.

        :rtype: ``List[:meth:`simplipy.system.SystemNotification`]``
        """
        return self._notifications

    @property  # type: ignore
    @guard_from_missing_data()
    def serial(self) -> str:
        """Return the system's serial number.

        :rtype: ``str``
        """
        return self._location_info["system"]["serial"]

    @property
    def state(self) -> SystemStates:
        """Return the current state of the system.

        :rtype: :meth:`simplipy.system.SystemStates`
        """
        return self._state

    @property  # type: ignore
    @guard_from_missing_data()
    def system_id(self) -> int:
        """Return the SimpliSafe identifier for this system.

        :rtype: ``int``
        """
        return self._location_info["sid"]

    @property  # type: ignore
    @guard_from_missing_data()
    def temperature(self) -> int:
        """Return the overall temperature measured by the system.

        :rtype: ``int``
        """
        return self._location_info["system"]["temperature"]

    @property  # type: ignore
    @guard_from_missing_data()
    def version(self) -> int:
        """Return the system version.

        :rtype: ``int``
        """
        return self._location_info["system"]["version"]

    @staticmethod
    def _coerce_state_from_string(value: str) -> SystemStates:
        """Return a proper state from a string input."""
        try:
            return SystemStates[convert_to_underscore(value)]
        except KeyError:
            _LOGGER.error("Unknown system state: %s", value)
            return SystemStates.unknown

    def _generate_system_notification_objects(self) -> List[SystemNotification]:
        """Generate message objects from the message data stored in location_info."""
        if self._location_info["system"].get("messages") is None:
            _LOGGER.info("Notifications unavailable in plan")
            return []

        return [
            SystemNotification(
                raw_message["id"],
                raw_message["text"],
                raw_message["category"],
                raw_message["code"],
                raw_message["timestamp"],
                link=raw_message["link"],
                link_label=raw_message["linkLabel"],
            )
            for raw_message in self._location_info["system"]["messages"]
        ]

    async def _get_entities(self, cached: bool = True) -> None:
        """Update sensors to the latest values."""
        entities: dict = await self._get_entities_payload(cached)

        _LOGGER.debug("Get entities response: %s", entities)

        entity_data: dict
        for entity_data in entities:
            if not entity_data:
                continue

            try:
                entity_type: EntityTypes = EntityTypes(entity_data["type"])
            except ValueError:
                _LOGGER.error("Unknown entity type: %s", entity_data["type"])
                entity_type = EntityTypes.unknown

            prop = self.locks if entity_type == EntityTypes.lock else self.sensors
            if entity_data["serial"] in prop:
                entity = prop[entity_data["serial"]]
                entity.entity_data = entity_data
            else:
                klass = get_entity_class(entity_type, version=self.version)
                prop[entity_data["serial"]] = klass(  # type: ignore
                    self._request,
                    self._get_entities,
                    self.system_id,
                    entity_type,
                    entity_data,
                )

    async def _get_entities_payload(self, cached: bool = False) -> dict:
        """Return the current sensor payload."""
        raise NotImplementedError()

    async def _get_settings(self, cached: bool = True) -> None:
        """Get all system settings."""
        pass

    async def _get_system_info(self) -> None:
        """Update information on the system."""
        subscription_resp: dict = await self._get_subscription_data()
        location_info: dict = next(
            (
                system["location"]
                for system in subscription_resp["subscriptions"]
                if system["sid"] == self.system_id
            )
        )

        self._location_info = location_info
        self._notifications = self._generate_system_notification_objects()
        self._state = self._coerce_state_from_string(
            location_info["system"]["alarmState"]
        )

    async def _set_updated_pins(self, pins: dict) -> None:
        """Post new PINs."""
        raise NotImplementedError()

    async def _set_state(self, value: SystemStates) -> None:
        """Raise if calling this undefined based method."""
        raise NotImplementedError()

    async def clear_notifications(self):
        """Clear all active notifications.

        This will remove the notification from SimpliSafe's cloud, meaning it will no
        longer visible in the SimpliSafe mobile and web apps.
        """
        if self._notifications:
            await self._request("delete", f"subscriptions/{self.system_id}/messages")
            self._notifications = []

    async def get_events(
        self, from_datetime: Optional[datetime] = None, num_events: Optional[int] = None
    ) -> list:
        """Get events recorded by the base station.

        If no parameters are provided, this will return the most recent 50 events.

        :param from_datetime: The starting datetime (if desired)
        :type from_datetime: ``datetime.datetime``
        :param num_events: The number of events to return.
        :type num_events: ``int``
        :rtype: ``list``
        """
        params: Dict[str, Any] = {}
        if from_datetime:
            params["fromTimestamp"] = round(from_datetime.timestamp())
        if num_events:
            params["numEvents"] = num_events

        events_resp: dict = await self._request(
            "get", f"subscriptions/{self.system_id}/events", params=params
        )

        _LOGGER.debug("Events response: %s", events_resp)

        return events_resp.get("events", [])

    async def get_latest_event(self) -> dict:
        """Get the most recent system event.

        :rtype: ``dict``
        """
        events: list = await self.get_events(num_events=1)

        try:
            return events[0]
        except IndexError:
            raise SimplipyError("SimpliSafe cloud didn't return any events")

    async def get_pins(self, cached: bool = True) -> Dict[str, str]:
        """Return all of the set PINs, including master and duress.

        The ``cached`` parameter determines whether the SimpliSafe Cloud uses the last
        known values retrieved from the base station (``True``) or retrieves new data.

        :param cached: Whether to used cached data.
        :type cached: ``bool``
        :rtype: ``Dict[str, str]``
        """
        raise NotImplementedError()

    async def remove_pin(self, pin_or_label: str) -> None:
        """Remove a PIN by its value or label.

        :param pin_or_label: The PIN value or label to remove
        :type pin_or_label: ``str``
        """
        # Because SimpliSafe's API works by sending the entire payload of PINs, we
        # can't reasonably check a local cache for up-to-date PIN data; so, we fetch the
        # latest each time:
        latest_pins: Dict[str, str] = await self.get_pins(cached=False)

        if pin_or_label in RESERVED_PIN_LABELS:
            raise PinError(f"Refusing to delete reserved PIN: {pin_or_label}")

        try:
            label: str = next(
                (k for k, v in latest_pins.items() if pin_or_label in (k, v))
            )
        except StopIteration:
            raise PinError(f"Cannot delete nonexistent PIN: {pin_or_label}")

        del latest_pins[label]

        await self._set_updated_pins(latest_pins)

    async def set_away(self) -> None:
        """Set the system in "Away" mode."""
        await self._set_state(SystemStates.away)

    async def set_home(self) -> None:
        """Set the system in "Home" mode."""
        await self._set_state(SystemStates.home)

    async def set_off(self) -> None:
        """Set the system in "Off" mode."""
        await self._set_state(SystemStates.off)

    async def set_pin(self, label: str, pin: str) -> None:
        """Set a PIN.

        :param label: The label to use for the PIN (shown in the SimpliSafe app)
        :type label: str
        :param pin: The pin value
        :type pin: str
        """
        if len(pin) != MAX_PIN_LENGTH:
            raise PinError(f"PINs must be {MAX_PIN_LENGTH} digits long")

        try:
            int(pin)
        except ValueError:
            raise PinError("PINs can only contain numbers")

        # Because SimpliSafe's API works by sending the entire payload of PINs, we
        # can't reasonably check a local cache for up-to-date PIN data; so, we fetch the
        # latest each time.
        latest_pins: Dict[str, str] = await self.get_pins(cached=False)

        if pin in latest_pins.values():
            raise PinError(f"Refusing to create duplicate PIN: {pin}")

        max_pins: int = DEFAULT_MAX_USER_PINS + len(RESERVED_PIN_LABELS)
        if len(latest_pins) == max_pins and label not in RESERVED_PIN_LABELS:
            raise PinError(f"Refusing to create more than {max_pins} user PINs")

        latest_pins[label] = pin

        await self._set_updated_pins(latest_pins)

    async def update(
        self,
        *,
        include_system: bool = True,
        include_settings: bool = True,
        include_entities: bool = True,
        cached: bool = True,
    ) -> None:
        """Get the latest system data.

        The ``cached`` parameter determines whether the SimpliSafe Cloud uses the last
        known values retrieved from the base station (``True``) or retrieves new data.

        :param include_system: Whether system state/properties should be updated
        :type include_system: ``bool``
        :param include_settings: Whether system settings (like PINs) should be updated
        :type include_settings: ``bool``
        :param include_entities: Whether sensors/locks/etc. should be updated
        :type include_entities: ``bool``
        :param cached: Whether to used cached data.
        :type cached: ``bool``
        """
        tasks: Dict[str, Coroutine] = {}
        if include_system:
            tasks["system"] = self._get_system_info()
        if include_settings:
            tasks["settings"] = self._get_settings(cached)

        results: List[Any] = await asyncio.gather(
            *tasks.values(), return_exceptions=True
        )

        operation: str
        result: dict
        for operation, result in zip(tasks, results):
            if isinstance(result, SimplipyError):
                raise result

        # We await entity updates after the task pool since including it can cause
        # HTTP 409s if that update occurs out of sequence:
        if include_entities:
            await self._get_entities(cached)



"""Define various string utilities."""
import re


def convert_to_underscore(string: str) -> str:
    """Convert thisString to this_string."""
    first_pass = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", string)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", first_pass).lower()



"""Define a base SimpliSafe entity."""
from enum import Enum
import logging
from typing import Callable, Coroutine

_LOGGER: logging.Logger = logging.getLogger(__name__)


class EntityTypes(Enum):
    """Entity types based on internal SimpliSafe ID number."""

    remote = 0
    keypad = 1
    keychain = 2
    panic_button = 3
    motion = 4
    entry = 5
    glass_break = 6
    carbon_monoxide = 7
    smoke = 8
    leak = 9
    temperature = 10
    camera = 12
    siren = 13
    doorbell = 15
    lock = 16
    lock_keypad = 253
    unknown = 99


class Entity:
    """A base SimpliSafe entity.

    Note that this class shouldn't be instantiated directly; it will be instantiated as
    appropriate via :meth:`simplipy.API.get_systems`.

    :param request: A method to make authenticated API requests.
    :type request: ``Callable[..., Coroutine]``
    :param update_func: A method to update the entity.
    :type update_func: ``Callable[..., Coroutine]``
    :param system_id: A SimpliSafe system ID.
    :type system_id: ``int``
    :param entity_type: The type of entity that this object represents.
    :type entity_type: ``simplipy.entity.EntityTypes``
    :param entity_data: A raw data dict representing the entity's state and properties.
    :type entity_data: ``dict``
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        request: Callable[..., Coroutine],
        update_func: Callable[..., Coroutine],
        system_id: int,
        entity_type: EntityTypes,
        entity_data: dict,
    ) -> None:
        """Initialize."""
        self._request: Callable[..., Coroutine] = request
        self._system_id: int = system_id
        self._type: EntityTypes = entity_type
        self._update_func: Callable[..., Coroutine] = update_func
        self.entity_data: dict = entity_data

    @property
    def name(self) -> str:
        """Return the entity name.

        :rtype: ``str``
        """
        return self.entity_data["name"]

    @property
    def serial(self) -> str:
        """Return the entity's serial number.

        :rtype: ``str``
        """
        return self.entity_data["serial"]

    @property
    def type(self) -> EntityTypes:
        """Return the entity type.

        :rtype: :meth:`simplipy.entity.EntityTypes`
        """
        return self._type

    async def update(self, cached: bool = True) -> None:
        """Retrieve the latest state/properties for the entity.

        The ``cached`` parameter determines whether the SimpliSafe Cloud uses the last
        known values retrieved from the base station (``True``) or retrieves new data.

        :param cached: Whether to used cached data.
        :type cached: ``bool``
        """
        await self._update_func(cached)


class EntityV3(Entity):
    """A base entity for V3 systems.

    Note that this class shouldn't be instantiated directly; it will be
    instantiated as appropriate via :meth:`simplipy.API.get_systems`.
    """

    @property
    def error(self) -> bool:
        """Return the entity's error status.

        :rtype: ``bool``
        """
        return self.entity_data["status"].get("malfunction", False)

    @property
    def low_battery(self) -> bool:
        """Return whether the entity's battery is low.

        :rtype: ``bool``
        """
        return self.entity_data["flags"]["lowBattery"]

    @property
    def offline(self) -> bool:
        """Return whether the entity is offline.

        :rtype: ``bool``
        """
        return self.entity_data["flags"]["offline"]

    @property
    def settings(self) -> dict:
        """Return the entity's settings.

        Note that these can change based on what entity type the entity is.

        :rtype: ``dict``
        """
        return self.entity_data["setting"]



"""Define a connection to the SimpliSafe websocket."""
import asyncio
from dataclasses import InitVar, dataclass, field
from datetime import datetime
import logging
from typing import Awaitable, Callable, Optional
from urllib.parse import urlencode

from socketio import AsyncClient
from socketio.exceptions import ConnectionError as ConnError, SocketIOError

from simplipy.entity import EntityTypes
from simplipy.errors import WebsocketError
from simplipy.util.dt import utc_from_timestamp

_LOGGER = logging.getLogger(__name__)

API_URL_BASE: str = "wss://api.simplisafe.com/socket.io"

DEFAULT_WATCHDOG_TIMEOUT = 900

EVENT_ALARM_CANCELED = "alarm_canceled"
EVENT_ALARM_TRIGGERED = "alarm_triggered"
EVENT_ARMED_AWAY = "armed_away"
EVENT_ARMED_AWAY_BY_KEYPAD = "armed_away_by_keypad"
EVENT_ARMED_AWAY_BY_REMOTE = "armed_away_by_remote"
EVENT_ARMED_HOME = "armed_home"
EVENT_AUTOMATIC_TEST = "automatic_test"
EVENT_AWAY_EXIT_DELAY_BY_KEYPAD = "away_exit_delay_by_keypad"
EVENT_AWAY_EXIT_DELAY_BY_REMOTE = "away_exit_delay_by_remote"
EVENT_CAMERA_MOTION_DETECTED = "camera_motion_detected"
EVENT_CONNECTION_LOST = "connection_lost"
EVENT_CONNECTION_RESTORED = "connection_restored"
EVENT_DISARMED_BY_MASTER_PIN = "disarmed_by_master_pin"
EVENT_DISARMED_BY_REMOTE = "disarmed_by_remote"
EVENT_DOORBELL_DETECTED = "doorbell_detected"
EVENT_ENTRY_DETECTED = "entry_detected"
EVENT_HOME_EXIT_DELAY = "home_exit_delay"
EVENT_LOCK_ERROR = "lock_error"
EVENT_LOCK_LOCKED = "lock_locked"
EVENT_LOCK_UNLOCKED = "lock_unlocked"
EVENT_MOTION_DETECTED = "motion_detected"
EVENT_POWER_OUTAGE = "power_outage"
EVENT_POWER_RESTORED = "power_restored"
EVENT_SENSOR_NOT_RESPONDING = "sensor_not_responding"
EVENT_SENSOR_RESTORED = "sensor_restored"

EVENT_MAPPING = {
    1110: EVENT_ALARM_TRIGGERED,
    1120: EVENT_ALARM_TRIGGERED,
    1132: EVENT_ALARM_TRIGGERED,
    1134: EVENT_ALARM_TRIGGERED,
    1154: EVENT_ALARM_TRIGGERED,
    1159: EVENT_ALARM_TRIGGERED,
    1162: EVENT_ALARM_TRIGGERED,
    1170: EVENT_CAMERA_MOTION_DETECTED,
    1301: EVENT_POWER_OUTAGE,
    1350: EVENT_CONNECTION_LOST,
    1381: EVENT_SENSOR_NOT_RESPONDING,
    1400: EVENT_DISARMED_BY_MASTER_PIN,
    1406: EVENT_ALARM_CANCELED,
    1407: EVENT_DISARMED_BY_REMOTE,
    1409: EVENT_MOTION_DETECTED,
    1429: EVENT_ENTRY_DETECTED,
    1458: EVENT_DOORBELL_DETECTED,
    1602: EVENT_AUTOMATIC_TEST,
    3301: EVENT_POWER_RESTORED,
    3350: EVENT_CONNECTION_RESTORED,
    3381: EVENT_SENSOR_RESTORED,
    3401: EVENT_ARMED_AWAY_BY_KEYPAD,
    3407: EVENT_ARMED_AWAY_BY_REMOTE,
    3441: EVENT_ARMED_HOME,
    3481: EVENT_ARMED_AWAY,
    3487: EVENT_ARMED_AWAY,
    3491: EVENT_ARMED_HOME,
    9401: EVENT_AWAY_EXIT_DELAY_BY_KEYPAD,
    9407: EVENT_AWAY_EXIT_DELAY_BY_REMOTE,
    9441: EVENT_HOME_EXIT_DELAY,
    9700: EVENT_LOCK_UNLOCKED,
    9701: EVENT_LOCK_LOCKED,
    9703: EVENT_LOCK_ERROR,
}


@dataclass(frozen=True)  # pylint: disable=too-many-instance-attributes
class WebsocketEvent:
    """Define a representation of a message."""

    event_cid: InitVar[int]
    info: str
    system_id: int
    timestamp: datetime

    event_type: Optional[str] = field(init=False)

    changed_by: Optional[str] = None
    sensor_name: Optional[str] = None
    sensor_serial: Optional[str] = None
    sensor_type: Optional[EntityTypes] = None

    def __post_init__(self, event_cid):
        """Run post-init initialization."""
        if event_cid in EVENT_MAPPING:
            object.__setattr__(self, "event_type", EVENT_MAPPING[event_cid])
        else:
            _LOGGER.warning(
                'Encountered unknown websocket event type: %s ("%s"). Please report it '
                "at https://github.com/bachya/simplisafe-python/issues.",
                event_cid,
                self.info,
            )
            object.__setattr__(self, "event_type", None)

        object.__setattr__(self, "timestamp", utc_from_timestamp(self.timestamp))

        if self.sensor_type is not None:
            try:
                object.__setattr__(self, "sensor_type", EntityTypes(self.sensor_type))
            except ValueError:
                _LOGGER.warning(
                    'Encountered unknown entity type: %s ("%s"). Please report it at'
                    "https://github.com/home-assistant/home-assistant/issues.",
                    self.sensor_type,
                    self.info,
                )
                object.__setattr__(self, "sensor_type", None)


def websocket_event_from_raw_data(event: dict):
    """Create a Message object from a websocket event payload."""
    return WebsocketEvent(
        event["eventCid"],
        event["info"],
        event["sid"],
        event["eventTimestamp"],
        changed_by=event["pinName"],
        sensor_name=event["sensorName"],
        sensor_serial=event["sensorSerial"],
        sensor_type=event["sensorType"],
    )


class WebsocketWatchdog:  # pylint: disable=too-few-public-methods
    """A watchdog that will ensure the websocket connection is functioning."""

    def __init__(
        self,
        action: Callable[..., Awaitable],
        *,
        timeout_seconds: int = DEFAULT_WATCHDOG_TIMEOUT,
    ):
        """Initialize."""
        self._action: Callable[..., Awaitable] = action
        self._loop = asyncio.get_event_loop()
        self._timer_task: Optional[asyncio.TimerHandle] = None
        self._timeout: int = timeout_seconds

    def cancel(self):
        """Cancel the watchdog."""
        if self._timer_task:
            self._timer_task.cancel()
            self._timer_task = None

    async def on_expire(self):
        """Log and act when the watchdog expires."""
        _LOGGER.info("Watchdog expired – calling %s", self._action.__name__)
        await self._action()

    async def trigger(self):
        """Trigger the watchdog."""
        _LOGGER.info("Watchdog triggered – sleeping for %s seconds", self._timeout)

        if self._timer_task:
            self._timer_task.cancel()

        self._timer_task = self._loop.call_later(
            self._timeout, lambda: asyncio.create_task(self.on_expire())
        )


class Websocket:
    """A websocket connection to the SimpliSafe cloud.

    Note that this class shouldn't be instantiated directly; it will be instantiated as
    appropriate via :meth:`simplipy.API.login_via_credentials` or
    :meth:`simplipy.API.login_via_token`.

    :param access_token: A SimpliSafe access token
    :type access_token: ``str``
    :param user_id: A SimpliSafe user ID
    :type user_id: ``int``
    """

    def __init__(self) -> None:
        """Initialize."""
        self._async_disconnect_handler: Optional[Callable[..., Awaitable]] = None
        self._sio: AsyncClient = AsyncClient()
        self._sync_disconnect_handler: Optional[Callable[..., Awaitable]] = None
        self._watchdog: WebsocketWatchdog = WebsocketWatchdog(self.async_reconnect)

        # Set by async_init():
        self._access_token: Optional[str] = None
        self._namespace: Optional[str] = None

    async def async_init(
        self, access_token: Optional[str], user_id: Optional[int] = None
    ) -> None:
        """Set the user ID and generate the namespace."""
        if not self._namespace:
            self._namespace = f"/v1/user/{user_id}"

        self._access_token = access_token

        # If the websocket is connected, reconnect it:
        if self._sio.connected:
            await self.async_reconnect()

    async def async_connect(self) -> None:
        """Connect to the socket."""
        params = {"ns": self._namespace, "accessToken": self._access_token}
        try:
            await self._sio.connect(
                f"{API_URL_BASE}?{urlencode(params)}",
                namespaces=[self._namespace],
                transports=["websocket"],
            )
        except (ConnError, SocketIOError) as err:
            raise WebsocketError(err) from None

    async def async_disconnect(self) -> None:
        """Disconnect from the socket."""
        await self._sio.disconnect()
        self._watchdog.cancel()

        if self._async_disconnect_handler:
            await self._async_disconnect_handler()
            self._async_disconnect_handler = None
        if self._sync_disconnect_handler:
            self._sync_disconnect_handler()
            self._sync_disconnect_handler = None

    def async_on_connect(self, target: Callable[..., Awaitable]) -> None:  # noqa: D202
        """Define a coroutine to be called when connecting.

        :param target: A coroutine
        :type target: ``Callable[..., Awaitable]``
        """

        async def _async_on_connect():
            """Act when connection occurs."""
            await self._watchdog.trigger()
            await target()

        self._sio.on("connect", _async_on_connect)

    def on_connect(self, target: Callable) -> None:  # noqa: D202
        """Define a synchronous method to be called when connecting.

        :param target: A synchronous function
        :type target: ``Callable``
        """

        async def _on_connect():
            """Act when connection occurs."""
            await self._watchdog.trigger()
            target()

        self._sio.on("connect", _on_connect)

    def async_on_disconnect(self, target: Callable[..., Awaitable]) -> None:
        """Define a coroutine to be called when disconnecting.

        :param target: A coroutine
        :type target: ``Callable[..., Awaitable]``
        """
        self._async_disconnect_handler = target

    def on_disconnect(self, target: Callable) -> None:
        """Define a synchronous method to be called when disconnecting.

        :param target: A synchronous function
        :type target: ``Callable``
        """
        self._sync_disconnect_handler = target

    def async_on_event(self, target: Callable[..., Awaitable]) -> None:  # noqa: D202
        """Define a coroutine to be called an event is received.

        The couroutine will have a ``data`` parameter that contains the raw data from
        the event.

        :param target: A coroutine
        :type target: ``Callable[..., Awaitable]``
        """

        async def _async_on_event(event_data: dict):
            """Act on the Message object."""
            await self._watchdog.trigger()
            message = websocket_event_from_raw_data(event_data)
            await target(message)

        self._sio.on("event", _async_on_event, namespace=self._namespace)

    def on_event(self, target: Callable) -> None:  # noqa: D202
        """Define a synchronous method to be called when an event is received.

        The method will have a ``data`` parameter that contains the raw data from the
        event.

        :param target: A synchronous function
        :type target: ``Callable``
        """

        async def _async_on_event(event_data: dict):
            """Act on the Message object."""
            await self._watchdog.trigger()
            message = websocket_event_from_raw_data(event_data)
            target(message)

        self._sio.on("event", _async_on_event, namespace=self._namespace)

    async def async_reconnect(self) -> None:
        """Reconnect the websocket connection."""
        await self.async_disconnect()
        await asyncio.sleep(1)
        await self.async_connect()



"""Define package errors."""


class SimplipyError(Exception):
    """A base error."""

    pass


class InvalidCredentialsError(SimplipyError):
    """An error related to invalid credentials."""

    pass


class PendingAuthorizationError(SimplipyError):
    """An error ralted to an unconfirmed multi-factor authentication."""

    pass


class PinError(SimplipyError):
    """An error related to invalid PINs or PIN operations."""

    pass


class RequestError(SimplipyError):
    """An error related to invalid requests."""

    pass


class WebsocketError(SimplipyError):
    """An error related to generic websocket errors."""

    pass



"""Define a V3 (new) SimpliSafe system."""
from enum import Enum
import logging
from typing import Dict

import voluptuous as vol

from simplipy.system import (
    CONF_DURESS_PIN,
    CONF_MASTER_PIN,
    System,
    create_pin_payload,
    guard_from_missing_data,
)

_LOGGER: logging.Logger = logging.getLogger(__name__)

CONF_ALARM_DURATION = "alarm_duration"
CONF_ALARM_VOLUME = "alarm_volume"
CONF_CHIME_VOLUME = "chime_volume"
CONF_ENTRY_DELAY_AWAY = "entry_delay_away"
CONF_ENTRY_DELAY_HOME = "entry_delay_home"
CONF_EXIT_DELAY_AWAY = "exit_delay_away"
CONF_EXIT_DELAY_HOME = "exit_delay_home"
CONF_LIGHT = "light"
CONF_VOICE_PROMPT_VOLUME = "voice_prompt_volume"

VOLUME_OFF = 0
VOLUME_LOW = 1
VOLUME_MEDIUM = 2
VOLUME_HIGH = 3
VOLUMES = [VOLUME_OFF, VOLUME_LOW, VOLUME_MEDIUM, VOLUME_HIGH]

SYSTEM_PROPERTIES_VALUE_MAP = {
    "alarm_duration": "alarmDuration",
    "alarm_volume": "alarmVolume",
    "chime_volume": "doorChime",
    "entry_delay_away": "entryDelayAway",
    "entry_delay_home": "entryDelayHome",
    "exit_delay_away": "exitDelayAway",
    "exit_delay_home": "exitDelayHome",
    "light": "light",
    "voice_prompt_volume": "voicePrompts",
}

SYSTEM_PROPERTIES_PAYLOAD_SCHEMA = vol.Schema(
    {
        vol.Optional(CONF_ALARM_DURATION): vol.All(
            vol.Coerce(int), vol.Range(min=30, max=480)
        ),
        vol.Optional(CONF_ALARM_VOLUME): vol.All(vol.Coerce(int), vol.In(VOLUMES)),
        vol.Optional(CONF_CHIME_VOLUME): vol.All(vol.Coerce(int), vol.In(VOLUMES)),
        vol.Optional(CONF_ENTRY_DELAY_AWAY): vol.All(
            vol.Coerce(int), vol.Range(min=30, max=255)
        ),
        vol.Optional(CONF_ENTRY_DELAY_HOME): vol.All(
            vol.Coerce(int), vol.Range(min=0, max=255)
        ),
        vol.Optional(CONF_EXIT_DELAY_AWAY): vol.All(
            vol.Coerce(int), vol.Range(min=45, max=255)
        ),
        vol.Optional(CONF_EXIT_DELAY_HOME): vol.All(
            vol.Coerce(int), vol.Range(min=0, max=255)
        ),
        vol.Optional(CONF_LIGHT): bool,
        vol.Optional(CONF_VOICE_PROMPT_VOLUME): vol.All(
            vol.Coerce(int), vol.In(VOLUMES)
        ),
    }
)


class SystemV3(System):
    """Define a V3 (new) system."""

    def __init__(self, request, get_subscription_data, location_info) -> None:
        """Initialize."""
        super().__init__(request, get_subscription_data, location_info)
        self.settings_info: dict = {}

    @property  # type: ignore
    @guard_from_missing_data()
    def alarm_duration(self) -> int:
        """Return the number of seconds an activated alarm will sound for.

        :rtype: ``int``
        """
        return self.settings_info["settings"]["normal"][
            SYSTEM_PROPERTIES_VALUE_MAP["alarm_duration"]
        ]

    @property  # type: ignore
    @guard_from_missing_data()
    def alarm_volume(self) -> int:
        """Return the volume level of the alarm.

        :rtype: ``int``
        """
        return int(
            self.settings_info["settings"]["normal"][
                SYSTEM_PROPERTIES_VALUE_MAP["alarm_volume"]
            ]
        )

    @property  # type: ignore
    @guard_from_missing_data()
    def battery_backup_power_level(self) -> int:
        """Return the power rating of the battery backup.

        :rtype: ``int``
        """
        return self.settings_info["basestationStatus"]["backupBattery"]

    @property  # type: ignore
    @guard_from_missing_data()
    def chime_volume(self) -> int:
        """Return the volume level of the door chime.

        :rtype: ``int``
        """
        return int(
            self.settings_info["settings"]["normal"][
                SYSTEM_PROPERTIES_VALUE_MAP["chime_volume"]
            ]
        )

    @property  # type: ignore
    @guard_from_missing_data()
    def entry_delay_away(self) -> int:
        """Return the number of seconds to delay when returning to an "away" alarm.

        :rtype: ``int``
        """
        return self.settings_info["settings"]["normal"][
            SYSTEM_PROPERTIES_VALUE_MAP["entry_delay_away"]
        ]

    @property  # type: ignore
    @guard_from_missing_data()
    def entry_delay_home(self) -> int:
        """Return the number of seconds to delay when returning to an "home" alarm.

        :rtype: ``int``
        """
        return self.settings_info["settings"]["normal"][
            SYSTEM_PROPERTIES_VALUE_MAP["entry_delay_home"]
        ]

    @property  # type: ignore
    @guard_from_missing_data()
    def exit_delay_away(self) -> int:
        """Return the number of seconds to delay when exiting an "away" alarm.

        :rtype: ``int``
        """
        return self.settings_info["settings"]["normal"][
            SYSTEM_PROPERTIES_VALUE_MAP["exit_delay_away"]
        ]

    @property  # type: ignore
    @guard_from_missing_data()
    def exit_delay_home(self) -> int:
        """Return the number of seconds to delay when exiting an "home" alarm.

        :rtype: ``int``
        """
        return self.settings_info["settings"]["normal"][
            SYSTEM_PROPERTIES_VALUE_MAP["exit_delay_home"]
        ]

    @property  # type: ignore
    @guard_from_missing_data()
    def gsm_strength(self) -> int:
        """Return the signal strength of the cell antenna.

        :rtype: ``int``
        """
        return self.settings_info["basestationStatus"]["gsmRssi"]

    @property  # type: ignore
    @guard_from_missing_data()
    def light(self) -> bool:
        """Return whether the base station light is on.

        :rtype: ``bool``
        """
        return self.settings_info["settings"]["normal"][
            SYSTEM_PROPERTIES_VALUE_MAP["light"]
        ]

    @property  # type: ignore
    @guard_from_missing_data(True)
    def offline(self) -> bool:
        """Return whether the system is offline.

        :rtype: ``bool``
        """
        return self._location_info["system"]["isOffline"]

    @property  # type: ignore
    @guard_from_missing_data(False)
    def power_outage(self) -> bool:
        """Return whether the system is experiencing a power outage.

        :rtype: ``bool``
        """
        return self._location_info["system"]["powerOutage"]

    @property  # type: ignore
    @guard_from_missing_data(False)
    def rf_jamming(self) -> bool:
        """Return whether the base station is noticing RF jamming.

        :rtype: ``bool``
        """
        return self.settings_info["basestationStatus"]["rfJamming"]

    @property  # type: ignore
    @guard_from_missing_data()
    def voice_prompt_volume(self) -> int:
        """Return the volume level of the voice prompt.

        :rtype: ``int``
        """
        return self.settings_info["settings"]["normal"][
            SYSTEM_PROPERTIES_VALUE_MAP["voice_prompt_volume"]
        ]

    @property  # type: ignore
    @guard_from_missing_data()
    def wall_power_level(self) -> int:
        """Return the power rating of the A/C outlet.

        :rtype: ``int``
        """
        return self.settings_info["basestationStatus"]["wallPower"]

    @property  # type: ignore
    @guard_from_missing_data()
    def wifi_ssid(self) -> str:
        """Return the ssid of the base station.

        :rtype: ``str``
        """
        return self.settings_info["settings"]["normal"]["wifiSSID"]

    @property  # type: ignore
    @guard_from_missing_data()
    def wifi_strength(self) -> int:
        """Return the signal strength of the wifi antenna.

        :rtype: ``int``
        """
        return self.settings_info["basestationStatus"]["wifiRssi"]

    async def _get_entities_payload(self, cached: bool = True) -> dict:
        """Update sensors to the latest values."""
        sensor_resp = await self._request(
            "get",
            f"ss3/subscriptions/{self.system_id}/sensors",
            params={"forceUpdate": str(not cached).lower()},
        )

        return sensor_resp.get("sensors", [])

    async def _get_settings(self, cached: bool = True) -> None:
        """Get all system settings."""
        settings_resp: dict = await self._request(
            "get",
            f"ss3/subscriptions/{self.system_id}/settings/normal",
            params={"forceUpdate": str(not cached).lower()},
        )

        if settings_resp:
            self.settings_info = settings_resp

    async def _set_state(self, value: Enum) -> None:
        """Set the state of the system."""
        state_resp: dict = await self._request(
            "post", f"ss3/subscriptions/{self.system_id}/state/{value.name}"
        )

        _LOGGER.debug('Set "%s" response: %s', value.name, state_resp)

        self._state = self._coerce_state_from_string(state_resp["state"])

    async def _set_updated_pins(self, pins: dict) -> None:
        """Post new PINs."""
        self.settings_info = await self._request(
            "post",
            f"ss3/subscriptions/{self.system_id}/settings/pins",
            json=create_pin_payload(pins),
        )

    async def get_pins(self, cached: bool = True) -> Dict[str, str]:
        """Return all of the set PINs, including master and duress.

        The ``cached`` parameter determines whether the SimpliSafe Cloud uses the last
        known values retrieved from the base station (``True``) or retrieves new data.

        :param cached: Whether to used cached data.
        :type cached: ``bool``
        :rtype: ``Dict[str, str]``
        """
        await self._get_settings(cached)

        pins: Dict[str, str] = {
            CONF_MASTER_PIN: self.settings_info["settings"]["pins"]["master"]["pin"],
            CONF_DURESS_PIN: self.settings_info["settings"]["pins"]["duress"]["pin"],
        }

        user_pin: dict
        for user_pin in [
            p for p in self.settings_info["settings"]["pins"]["users"] if p["pin"]
        ]:
            pins[user_pin["name"]] = user_pin["pin"]

        return pins

    async def set_properties(self, properties: dict) -> None:
        """Set various system properties.

        The following properties can be set:
           1. alarm_duration (in seconds): 30-480
           2. alarm_volume: 0 (off), 1 (low), 2 (medium), 3 (high)
           3. chime_volume: 0 (off), 1 (low), 2 (medium), 3 (high)
           4. entry_delay_away (in seconds): 30-255
           5. entry_delay_home (in seconds): 0-255
           6. exit_delay_away (in seconds): 45-255
           7. exit_delay_home (in seconds): 0-255
           8. light: True or False
           9. voice_prompt_volume: 0 (off), 1 (low), 2 (medium), 3 (high)

        :param properties: The system properties to set.
        :type properties: ``dict``
        """
        try:
            SYSTEM_PROPERTIES_PAYLOAD_SCHEMA(properties)
        except vol.Invalid as err:
            raise ValueError(
                f"Using invalid values for system properties ({properties}): {err}"
            ) from None

        settings_resp = await self._request(
            "post",
            f"ss3/subscriptions/{self.system_id}/settings/normal",
            json={
                "normal": {
                    SYSTEM_PROPERTIES_VALUE_MAP[prop]: value
                    for prop, value in properties.items()
                }
            },
        )

        if settings_resp:
            self.settings_info = settings_resp



"""Define a v2 (old) SimpliSafe sensor."""
import logging

from simplipy.entity import Entity, EntityTypes
from simplipy.errors import SimplipyError

_LOGGER: logging.Logger = logging.getLogger(__name__)


class SensorV2(Entity):
    """A V2 (old) sensor.

    Note that this class shouldn't be instantiated directly; it will be
    instantiated as appropriate via :meth:`simplipy.API.get_systems`.
    """

    @property
    def data(self) -> int:
        """Return the sensor's current data flag (currently not understood).

        :rtype: ``int``
        """
        return self.entity_data["sensorData"]

    @property
    def error(self) -> bool:
        """Return the sensor's error status.

        :rtype: ``bool``
        """
        return self.entity_data["error"]

    @property
    def low_battery(self) -> bool:
        """Return whether the sensor's battery is low.

        :rtype: ``bool``
        """
        return self.entity_data.get("battery", "ok") != "ok"

    @property
    def settings(self) -> bool:
        """Return the sensor's settings.

        :rtype: ``bool``
        """
        return self.entity_data["setting"]

    @property
    def trigger_instantly(self) -> bool:
        """Return whether the sensor will trigger instantly.

        :rtype: ``bool``
        """
        return self.entity_data["instant"]

    @property
    def triggered(self) -> bool:
        """Return whether the sensor has been triggered.

        :rtype: ``bool``
        """
        if self.type == EntityTypes.entry:
            return self.entity_data.get("entryStatus", "closed") == "open"

        raise SimplipyError(f"Cannot determine triggered state for sensor: {self.name}")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['simplipy', 'simplipy.sensor', 'simplipy.system', 'simplipy.util']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.6.2,<4.0.0',
 'python-engineio==3.13.1',
 'python-socketio==4.6.0',
 'pytz>=2019.3,<2021.0',
 'voluptuous>=0.11.7,<0.12.0',
 'websockets>=8.1,<9.0']

setup_kwargs = {
    'name': 'simplisafe-python',
    'version': '9.2.1',
    'description': 'A Python3, async interface to the SimpliSafe API',
    'long_description': '# 🚨 simplisafe-python: A Python3, async interface to the SimpliSafe™ API\n\n[![CI](https://github.com/bachya/simplisafe-python/workflows/CI/badge.svg)](https://github.com/bachya/simplisafe-python/actions)\n[![PyPi](https://img.shields.io/pypi/v/simplisafe-python.svg)](https://pypi.python.org/pypi/simplisafe-python)\n[![Version](https://img.shields.io/pypi/pyversions/simplisafe-python.svg)](https://pypi.python.org/pypi/simplisafe-python)\n[![License](https://img.shields.io/pypi/l/simplisafe-python.svg)](https://github.com/bachya/simplisafe-python/blob/master/LICENSE)\n[![Code Coverage](https://codecov.io/gh/bachya/simplisafe-python/branch/master/graph/badge.svg)](https://codecov.io/gh/bachya/simplisafe-python)\n[![Maintainability](https://api.codeclimate.com/v1/badges/f46d8b1dcfde6a2f683d/maintainability)](https://codeclimate.com/github/bachya/simplisafe-python/maintainability)\n[![Say Thanks](https://img.shields.io/badge/SayThanks-!-1EAEDB.svg)](https://saythanks.io/to/bachya)\n\n`simplisafe-python` (hereafter referred to as `simplipy`) is a Python3,\n`asyncio`-driven interface to the unofficial SimpliSafe™ API. With it, users can\nget data on their system (including available sensors), set the system state,\nand more.\n\n# Documentation\n\nYou can find complete documentation here: https://simplisafe-python.readthedocs.io\n\n# Contributing\n\n1. [Check for open features/bugs](https://github.com/bachya/simplisafe-python/issues)\n  or [initiate a discussion on one](https://github.com/bachya/simplisafe-python/issues/new).\n2. [Fork the repository](https://github.com/bachya/simplisafe-python/fork).\n3. (_optional, but highly recommended_) Create a virtual environment: `python3 -m venv .venv`\n4. (_optional, but highly recommended_) Enter the virtual environment: `source ./.venv/bin/activate`\n5. Install the dev environment: `script/setup`\n6. Code your new feature or bug fix.\n7. Write tests that cover your new functionality.\n8. Run tests and ensure 100% code coverage: `script/test`\n9. Update `README.md` with any new documentation.\n10. Add yourself to `AUTHORS.md`.\n11. Submit a pull request!\n',
    'author': 'Aaron Bach',
    'author_email': 'bachya1208@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bachya/simplisafe-python',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.0,<4.0.0',
}


setup(**setup_kwargs)



"""Define a V2 (original) SimpliSafe system."""
from enum import Enum
import logging
from typing import Dict

from simplipy.system import (
    CONF_DURESS_PIN,
    CONF_MASTER_PIN,
    System,
    SystemStates,
    create_pin_payload,
)

_LOGGER: logging.Logger = logging.getLogger(__name__)


class SystemV2(System):
    """Define a V2 (original) system."""

    async def _get_entities_payload(self, cached: bool = True) -> dict:
        """Update sensors to the latest values."""
        sensor_resp = await self._request(
            "get",
            f"subscriptions/{self.system_id}/settings",
            params={"settingsType": "all", "cached": str(cached).lower()},
        )

        return sensor_resp.get("settings", {}).get("sensors", [])

    async def _set_updated_pins(self, pins: dict) -> None:
        """Post new PINs."""
        await self._request(
            "post",
            f"subscriptions/{self.system_id}/pins",
            json=create_pin_payload(pins, version=2),
        )

    async def _set_state(self, value: Enum) -> None:
        """Set the state of the system."""
        state_resp: dict = await self._request(
            "post",
            f"subscriptions/{self.system_id}/state",
            params={"state": value.name},
        )

        _LOGGER.debug('Set "%s" response: %s', value.name, state_resp)

        if state_resp["success"]:
            self._state = SystemStates[state_resp["requestedState"]]

    async def get_pins(self, cached: bool = True) -> Dict[str, str]:
        """Return all of the set PINs, including master and duress.

        The ``cached`` parameter determines whether the SimpliSafe Cloud uses the last
        known values retrieved from the base station (``True``) or retrieves new data.

        :param cached: Whether to used cached data.
        :type cached: ``bool``
        :rtype: ``Dict[str, str]``
        """
        pins_resp: dict = await self._request(
            "get",
            f"subscriptions/{self.system_id}/pins",
            params={"settingsType": "all", "cached": str(cached).lower()},
        )

        pins: Dict[str, str] = {
            CONF_MASTER_PIN: pins_resp["pins"].pop("pin1")["value"],
            CONF_DURESS_PIN: pins_resp["pins"].pop("duress")["value"],
        }

        user_pin: dict
        for user_pin in [p for p in pins_resp["pins"].values() if p["value"]]:
            pins[user_pin["name"]] = user_pin["value"]

        return pins



"""Define a SimpliSafe account."""
import base64
from datetime import datetime, timedelta
from json.decoder import JSONDecodeError
import logging
from typing import Dict, Optional, Type, TypeVar
from uuid import uuid4

from aiohttp import ClientSession, ClientTimeout
from aiohttp.client_exceptions import ClientError

from simplipy.errors import (
    InvalidCredentialsError,
    PendingAuthorizationError,
    RequestError,
)
from simplipy.system import System
from simplipy.system.v2 import SystemV2
from simplipy.system.v3 import SystemV3
from simplipy.websocket import Websocket

_LOGGER: logging.Logger = logging.getLogger(__name__)

API_URL_HOSTNAME: str = "api.simplisafe.com"
API_URL_BASE: str = f"https://{API_URL_HOSTNAME}/v1"
API_URL_MFA_OOB: str = "http://simplisafe.com/oauth/grant-type/mfa-oob"

DEFAULT_TIMEOUT: int = 10
DEFAULT_USER_AGENT: str = (
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_6) AppleWebKit/605.1.15 "
    "(KHTML, like Gecko) Version/13.1.2 Safari/605.1.15"
)

CLIENT_ID_TEMPLATE = "{0}.WebApp.simplisafe.com"
DEVICE_ID_TEMPLATE = (
    'WebApp; useragent="Safari 13.1 (SS-ID: {0}) / macOS 10.15.6"; uuid="{1}"; id="{0}"'
)

SYSTEM_MAP: Dict[int, Type[System]] = {2: SystemV2, 3: SystemV3}


ApiType = TypeVar("ApiType", bound="API")


def generate_device_id(client_id: str) -> str:
    """Generate a random 10-character ID to use as the SimpliSafe device ID."""
    seed = base64.b64encode(client_id.encode()).decode()[:10]
    return f"{seed[:5]}-{seed[5:]}"


class API:  # pylint: disable=too-many-instance-attributes
    """An API object to interact with the SimpliSafe cloud.

    Note that this class shouldn't be instantiated directly; instead, the
    :meth:`simplipy.API.login_via_credentials` and :meth:`simplipy.API.login_via_token`
    class methods should be used.

    :param session: The ``aiohttp`` ``ClientSession`` session used for all HTTP requests
    :type session: ``aiohttp.client.ClientSession``
    :param client_id: The SimpliSafe client ID to use for this API object
    :type client_id: ``str``
    """

    def __init__(
        self, *, client_id: str, session: Optional[ClientSession] = None
    ) -> None:
        """Initialize."""
        self._access_token: Optional[str] = None
        self._access_token_expire: Optional[datetime] = None
        self._actively_refreshing: bool = False
        self._refresh_token: Optional[str] = None
        self._session: ClientSession = session

        self._client_id = client_id if client_id else str(uuid4())
        self._client_id_string: str = CLIENT_ID_TEMPLATE.format(self._client_id)
        self._device_id: str = generate_device_id(self._client_id)

        self.email: Optional[str] = None
        self.user_id: Optional[int] = None
        self.websocket: Websocket = Websocket()

    @property
    def access_token(self) -> Optional[str]:
        """Return the current access token.

        :rtype: ``str``
        """
        return self._access_token

    @property
    def client_id(self) -> str:
        """Return the client ID of the API."""
        return self._client_id

    @property
    def client_id_string(self) -> str:
        """Return the client ID of the API."""
        return self._client_id_string

    @property
    def device_id(self) -> str:
        """Return the generated device ID for this API instance."""
        return self._device_id

    @property
    def refresh_token(self) -> Optional[str]:
        """Return the current refresh token.

        :rtype: ``str``
        """
        return self._refresh_token

    @classmethod
    async def login_via_credentials(
        cls: Type[ApiType],
        email: str,
        password: str,
        *,
        client_id: str,
        session: Optional[ClientSession] = None,
    ) -> ApiType:
        """Create an API object from a email address and password.

        :param email: A SimpliSafe email address
        :type email: ``str``
        :param password: A SimpliSafe password
        :type password: ``str``
        :param session: An ``aiohttp`` ``ClientSession``
        :type session: ``aiohttp.client.ClientSession``
        :param client_id: The SimpliSafe client ID to use for this API object
        :type client_id: ``str``
        :rtype: :meth:`simplipy.API`
        """
        klass = cls(session=session, client_id=client_id)
        klass.email = email

        await klass.authenticate(
            {
                "grant_type": "password",
                "username": email,
                "password": password,
                "client_id": klass.client_id_string,
                "device_id": DEVICE_ID_TEMPLATE.format(
                    klass.device_id, klass.client_id
                ),
                "app_version": "1.62.0",
                "scope": "offline_access",
            }
        )

        return klass

    @classmethod
    async def login_via_token(
        cls: Type[ApiType],
        refresh_token: str,
        *,
        client_id: str,
        session: Optional[ClientSession] = None,
    ) -> ApiType:
        """Create an API object from a refresh token.

        :param refresh_token: A SimpliSafe refresh token
        :type refresh_token: ``str``
        :param session: An ``aiohttp`` ``ClientSession``
        :type session: ``aiohttp.client.ClientSession``
        :param client_id: The SimpliSafe client ID to use for this API object
        :type client_id: ``str``
        :rtype: :meth:`simplipy.API`
        """
        klass = cls(session=session, client_id=client_id)
        await klass.refresh_access_token(refresh_token)
        return klass

    async def _get_subscription_data(self) -> dict:
        """Get the latest location-level data."""
        subscription_resp: dict = await self.request(
            "get", f"users/{self.user_id}/subscriptions", params={"activeOnly": "true"}
        )

        _LOGGER.debug(
            "users/%s/subscriptions response: %s", self.user_id, subscription_resp
        )

        return subscription_resp

    async def authenticate(self, payload: dict) -> None:
        """Authenticate the API object using an authentication payload."""
        _LOGGER.debug("Authentication payload: %s", payload)

        token_resp: dict = await self.request("post", "api/token", json=payload)

        if "mfa_token" in token_resp:
            mfa_challenge_response: dict = await self.request(
                "post",
                "api/mfa/challenge",
                json={
                    "challenge_type": "oob",
                    "client_id": self._client_id_string,
                    "mfa_token": token_resp["mfa_token"],
                },
            )

            await self.request(
                "post",
                "api/token",
                json={
                    "client_id": self._client_id_string,
                    "grant_type": API_URL_MFA_OOB,
                    "mfa_token": token_resp["mfa_token"],
                    "oob_code": mfa_challenge_response["oob_code"],
                    "scope": "offline_access",
                },
            )

            raise PendingAuthorizationError(
                f"Check your email for an MFA link, then use {self._client_id} "
                "as the client_id parameter in future API calls"
            )

        self._access_token = token_resp["access_token"]
        self._access_token_expire = datetime.now() + timedelta(
            seconds=int(token_resp["expires_in"]) - 60
        )
        self._refresh_token = token_resp["refresh_token"]

        auth_check_resp: dict = await self.request("get", "api/authCheck")
        self.user_id = auth_check_resp["userId"]

        await self.websocket.async_init(self._access_token, self.user_id)

    async def get_systems(self) -> Dict[str, System]:
        """Get systems associated to the associated SimpliSafe account.

        In the dict that is returned, the keys are the system ID and the values are
        actual ``System`` objects.

        :rtype: ``Dict[str, simplipy.system.System]``
        """
        subscription_resp: dict = await self._get_subscription_data()

        systems: Dict[str, System] = {}
        for system_data in subscription_resp["subscriptions"]:
            if "version" not in system_data["location"]["system"]:
                _LOGGER.error(
                    "Skipping location with missing system data: %s",
                    system_data["location"]["sid"],
                )
                continue

            version = system_data["location"]["system"]["version"]
            system_class = SYSTEM_MAP[version]
            system = system_class(
                self.request, self._get_subscription_data, system_data["location"]
            )
            await system.update(include_system=False)
            systems[system_data["sid"]] = system

        return systems

    async def request(  # pylint: disable=too-many-branches
        self, method: str, endpoint: str, **kwargs
    ) -> dict:
        """Make an API request."""
        if (
            self._access_token_expire
            and datetime.now() >= self._access_token_expire
            and not self._actively_refreshing
        ):
            _LOGGER.debug(
                "Need to refresh access token (expiration: %s)",
                self._access_token_expire,
            )
            self._actively_refreshing = True
            await self.refresh_access_token(self._refresh_token)

        kwargs.setdefault("headers", {})
        if self._access_token:
            kwargs["headers"]["Authorization"] = f"Bearer {self._access_token}"
        kwargs["headers"]["Content-Type"] = "application/json; charset=utf-8"
        kwargs["headers"]["Host"] = API_URL_HOSTNAME
        kwargs["headers"]["User-Agent"] = DEFAULT_USER_AGENT

        use_running_session = self._session and not self._session.closed

        if use_running_session:
            session = self._session
        else:
            session = ClientSession(timeout=ClientTimeout(total=DEFAULT_TIMEOUT))

        async with session.request(
            method, f"{API_URL_BASE}/{endpoint}", **kwargs
        ) as resp:
            try:
                data = await resp.json(content_type=None)
            except JSONDecodeError:
                message = await resp.text()
                data = {"error": message}

            _LOGGER.debug("Data received from /%s: %s", endpoint, data)

            try:
                resp.raise_for_status()
            except ClientError as err:
                if isinstance(data, dict) and data.get("error") == "mfa_required":
                    # In the case of an MFA token, SimpliSafe's API will return a 401,
                    # but will include the MFA token in the response body. This
                    # somewhat-kludgy check handles that case by ignoring the 401 and
                    # returning the body for use:
                    return data

                if "401" in str(err):
                    if self._actively_refreshing:
                        raise InvalidCredentialsError(
                            "Repeated 401s despite refreshing access token"
                        ) from None
                    if self._refresh_token:
                        _LOGGER.info("401 detected; attempting refresh token")
                        self._access_token_expire = datetime.now()
                        return await self.request(method, endpoint, **kwargs)
                    raise InvalidCredentialsError("Invalid username/password") from None

                if "403" in str(err):
                    raise InvalidCredentialsError(
                        f"User does not have permission to access {endpoint}"
                    ) from None

                raise RequestError(
                    f"There was an error while requesting /{endpoint}: {err}"
                ) from None
            finally:
                if not use_running_session:
                    await session.close()

        return data

    async def refresh_access_token(self, refresh_token: Optional[str]) -> None:
        """Regenerate an access token.

        :param refresh_token: The refresh token to use
        :type refresh_token: str
        """
        await self.authenticate(
            {
                "grant_type": "refresh_token",
                "client_id": self._client_id,
                "refresh_token": refresh_token,
            }
        )

        self._actively_refreshing = False
